@@ -1,4 +1,13 @@
-"""Tools for serialization and deserialization.
+"""Tools for serialization and deserialization by Andie
+
+Ctor short for constructor.
+
+The two main methods to be aware of are Ctor.deconstruct() and Ctor.construct(). Ideally, that's all
+you need to use to break down any object into an objest ready for json.dumps() and reconstruct back
+to an equivalent object.
+
+Include to_jdict() and from_jdict() for an easy way to make a custom class instance work with this.
+Otherwise, it will attempt to pickle any unhandled types.
 
 An example with numpy arrays:
 
@@ -9,7 +18,7 @@ An example with numpy arrays:
 >>> a_recon_from_jdict = Ctor.construct(a_decon_jdict)
 >>> assert np.allclose(a, a_recon_from_ctor_dict, a_recon_from_jdict)
 
-Another example (not runnable, since dependent on some third party code
+Another example not runnable, since dependent on some third party code
 
 ```
 from ... .source.audio import PyAudioSourceReader
@@ -20,19 +29,151 @@ c_decon_jdict = Ctor.deconstruct(c, validate_conversion=True, output_type=Ctor.J
 c_recon_from_ctor_dict = Ctor.construct(c_decon_ctor_dict)
 c_recon_from_jdict = Ctor.construct(c_decon_jdict)
 assert c == c_recon_from_jdict == c_recon_from_ctor_dict
-```
 
+The Design
+
+`Ctor.deconstruct()` will traverse through objects supported by remap (list, tuple, dict, set)
+to find objects foreign to json and deconstruct them using `Ctor.deconstruction_specs`.
+Specified object are deconstructed into a CtorDict containing CONSTRUCTOR, ARGS, and KWARGS.
+
+Converting a function to a CtorDict results in the CONSTRUCTOR as the Ctor deserializing method
+and the ARGS as the import path of the deconstructed function.
+
+>>> from pprint import pprint
+>>> from py2json.tests.test_ctor import add
+>>> serialized = Ctor.deconstruct(add, validate_conversion=True, output_type=Ctor.CTOR_DICT)
+>>> pprint(serialized, sort_dicts=False)
+{'CONSTRUCTOR': <bound method Ctor._deserialize_ctor_from_jdict of <class 'ctor.Ctor'>>,
+ 'ARGS': [{'module': 'py2json.tests.test_ctor', 'name': 'add', 'attr': None}],
+ 'KWARGS': None}
+
+By default, the output type is a json dict. The CONSTRUCTOR method is reduced to an import path.
+
+>>> serialized = Ctor.deconstruct(add, validate_conversion=True, output_type=Ctor.JSON_DICT)
+>>> pprint(serialized, sort_dicts=False)
+{'CONSTRUCTOR': {'module': 'ctor',
+                 'name': 'Ctor',
+                 'attr': '_deserialize_ctor_from_jdict'},
+ 'ARGS': [{'module': 'py2json.tests.test_ctor', 'name': 'add', 'attr': None}],
+ 'KWARGS': None}
+
+Ideally, a custom class will include `to_jdict` and `from_jdict` methods
+
+>>> from py2json.tests.test_ctor import TestToJdict
+>>> instance = TestToJdict('test_value')
+>>> serialized = Ctor.deconstruct(instance, validate_conversion=True)
+>>> pprint(serialized, sort_dicts=False)
+{'CONSTRUCTOR': {'module': 'py2json.tests.test_ctor',
+                 'name': 'TestToJdict',
+                 'attr': 'from_jdict'},
+ 'ARGS': [{'value': 'test_value'}],
+ 'KWARGS': None}
+
+As a fallback, dill is used when no matching deconstruction_specs is found.
+`CtorException` is thrown if it fails.
+
+>>> from py2json.tests.test_ctor import TestClass
+>>> instance = TestClass('test_value')
+>>> serialized = Ctor.deconstruct(instance, validate_conversion=True)
+>>> pprint(serialized, sort_dicts=False) # doctest: +SKIP
+{'CONSTRUCTOR': {'module': 'ctor', 'name': 'dill_load_string', 'attr': None},
+ 'ARGS': ['\x80\x04\x95E\x00\x00\x00\x00\x00\x00\x00\x8c\x17py2json.tests.test_ctor\x94\x8c\t'
+          'TestClass\x94\x93\x94)\x81\x94}\x94\x8c\x05value\x94\x8c\n'
+          'test_value\x94sb.'],
+ 'KWARGS': None}
+
+
+Ctor.deconstruct will also check if the deconstructed args need to be deconstructed as well.
+In this example, the object deconstructs with ARGS that also contain to_jdict/from_jdict
+
+>>> from py2json.tests.test_ctor import TestToJdict
+>>> instance = TestToJdict([TestToJdict('test_value')])
+>>> serialized = Ctor.deconstruct(instance, validate_conversion=True)
+>>> pprint(serialized, sort_dicts=False)
+{'CONSTRUCTOR': {'module': 'py2json.tests.test_ctor',
+                 'name': 'TestToJdict',
+                 'attr': 'from_jdict'},
+ 'ARGS': [{'value': [{'CONSTRUCTOR': {'module': 'py2json.tests.test_ctor',
+                                      'name': 'TestToJdict',
+                                      'attr': 'from_jdict'},
+                      'ARGS': [{'value': 'test_value'}],
+                      'KWARGS': None}]}],
+ 'KWARGS': None}
+
+In this example, the partial function deconstructs with ARGS that also contain a function
+
+>>> from functools import partial
+>>> partial_func = partial(add, 1, b=2)
+>>> serialized = Ctor.deconstruct(partial_func, validate_conversion=True, output_type=Ctor.CTOR_DICT)
+>>> pprint(serialized, sort_dicts=False)
+{'CONSTRUCTOR': <class 'functools.partial'>,
+ 'ARGS': [{'CONSTRUCTOR': {'module': 'ctor',
+                           'name': 'Ctor',
+                           'attr': '_deserialize_ctor_from_jdict'},
+           'ARGS': [{'module': 'py2json.tests.test_ctor',
+                     'name': 'add',
+                     'attr': None}],
+           'KWARGS': None},
+          1],
+ 'KWARGS': {'b': 2}}
+
+
+Function calls can also be serialized
+
+>>> import os
+>>> ctor_jdict = Ctor.to_ctor_dict(constructor=os.path.join, args=['I', 'am', 'a', 'filepath'])
+>>> pprint(ctor_jdict, sort_dicts=False)
+{'CONSTRUCTOR': {'module': 'posixpath', 'name': 'join', 'attr': None},
+ 'ARGS': ['I', 'am', 'a', 'filepath'],
+ 'KWARGS': {}}
+>>> Ctor.construct(ctor_jdict)
+'I/am/a/filepath'
+
+You can pass args
+>>> from collections import namedtuple
+>>> ctor_jdict = Ctor.to_ctor_dict(namedtuple, args=('A', 'x y z'))
+>>> pprint(ctor_jdict, sort_dicts=False)
+{'CONSTRUCTOR': {'module': 'collections', 'name': 'namedtuple', 'attr': None},
+ 'ARGS': ('A', 'x y z'),
+ 'KWARGS': {}}
+>>> A = Ctor.construct(ctor_jdict)
+>>> A('no', 'defaults', 'here')
+A(x='no', y='defaults', z='here')
+
+...or kwargs
+>>> ctor_jdict = Ctor.to_ctor_dict(namedtuple, kwargs={'typename': 'A', 'field_names': 'x y z'})
+>>> pprint(ctor_jdict, sort_dicts=False)
+{'CONSTRUCTOR': {'module': 'collections', 'name': 'namedtuple', 'attr': None},
+ 'ARGS': [],
+ 'KWARGS': {'typename': 'A', 'field_names': 'x y z'}}
+>>> A = Ctor.construct(ctor_jdict)
+>>> A('no', 'defaults', 'here')
+A(x='no', y='defaults', z='here')
+
+...or both args and kwargs
+>>> ctor_jdict = Ctor.to_ctor_dict(namedtuple, ('A', 'x y z'), {'defaults': ('has', 'defaults')})
+>>> pprint(ctor_jdict, sort_dicts=False)
+{'CONSTRUCTOR': {'module': 'collections', 'name': 'namedtuple', 'attr': None},
+ 'ARGS': ('A', 'x y z'),
+ 'KWARGS': {'defaults': ('has', 'defaults')}}
+>>> A = Ctor.construct(ctor_jdict)
+>>> A('this one')
+A(x='this one', y='has', z='defaults')
 
 """
 import functools
 import importlib
 import inspect
 import json
-from typing import Callable
+from typing import Callable, TypedDict, Union, Any, List, Dict
 
+import dill
 import numpy as np
 from boltons.iterutils import remap, default_enter
 from glom import Literal, glom, Spec
+from i2 import mk_sentinel
+
+NOT_SET = mk_sentinel('NOT_SET')
 
 
 def mk_serializer_and_deserializer(spec, mk_inv_spec=None):
@@ -138,6 +279,14 @@ ctor_to_jdict_spec = {
 )
 
 
+def dill_dump_string(x):
+    return dill.dumps(x).decode('latin-1')
+
+
+def dill_load_string(x):
+    return dill.loads(x.encode('latin-1'))
+
+
 class CtorException(ValueError):
     pass
 
@@ -154,6 +303,22 @@ class CtorNames:
     CONSTRUCTED = 'constructed'
 
 
+class ConstructorJdict(TypedDict):
+    module: str
+    name: str
+    attr: str
+
+
+CtorDict = TypedDict(
+    'CtorDict',
+    {
+        CtorNames.CONSTRUCTOR: Union[Callable, ConstructorJdict],
+        CtorNames.ARGS: List[Any],
+        CtorNames.KWARGS: Dict[str, Any],
+    },
+)
+
+
 class Ctor(CtorNames):
     """
     A Base class for serializing any object
@@ -167,6 +332,22 @@ class Ctor(CtorNames):
     def _deserialize_ctor_from_jdict(cls, ctor_jdict):
         return _deserialize_ctor_from_jdict(ctor_jdict)
 
+    @classproperty
+    def fallback_deconstruction_spec(cls):
+        """Try to pickle when no deconstruction spec is found"""
+        return {
+            'description': 'default to dill pickle',
+            'check_type': lambda x: True,
+            'spec': {
+                Ctor.CONSTRUCTOR: Literal(dill_load_string),
+                Ctor.ARGS: lambda x: [dill_dump_string(x)],
+                Ctor.KWARGS: Literal(None),
+            },
+            'validate_conversion': lambda x, serialized_x: (
+                x == Ctor._construct_obj(serialized_x)
+            ),
+        }
+
     @classproperty
     def deconstruction_specs(cls):
         """
@@ -176,6 +357,22 @@ class Ctor(CtorNames):
                                                    validate_conversion='optional callable comparing obj to ctor_dict')
         """
         return [
+            {
+                'description': 'For functools.partial',
+                'check_type': lambda x: isinstance(x, functools.partial),
+                'spec': {
+                    Ctor.CONSTRUCTOR: Literal(functools.partial),
+                    Ctor.ARGS: lambda x: [x.func, *x.args],
+                    Ctor.KWARGS: lambda x: dict(x.keywords),
+                },
+                'validate_conversion': lambda x, serialized_x: (
+                    (deserialized_x := Ctor._construct_obj(serialized_x))
+                    and all(
+                        getattr(x, a, NOT_SET) == getattr(deserialized_x, a, NOT_SET)
+                        for a in ('func', 'args', 'keywords')
+                    )
+                ),
+            },
             {
                 'description': 'For numpy.ndarray',
                 'check_type': lambda x: isinstance(x, np.ndarray),
@@ -184,8 +381,8 @@ class Ctor(CtorNames):
                     Ctor.ARGS: lambda x: [x.tolist()],
                     Ctor.KWARGS: Literal(None),
                 },
-                'validate_conversion': lambda x, serialized_x: np.allclose(
-                    x, Ctor._construct_obj(serialized_x)
+                'validate_conversion': lambda x, serialized_x: (
+                    np.allclose(x, Ctor._construct_obj(serialized_x))
                 ),
             },
             {
@@ -197,8 +394,9 @@ class Ctor(CtorNames):
                     Ctor.ARGS: lambda x: [x.to_jdict()],
                     Ctor.KWARGS: Literal(None),
                 },
-                'validate_conversion': lambda x, serialized_x: x.to_jdict()
-                == Ctor._construct_obj(serialized_x).to_jdict(),
+                'validate_conversion': lambda x, serialized_x: (
+                    x.to_jdict() == Ctor._construct_obj(serialized_x).to_jdict()
+                ),
             },
             {
                 'description': 'for class methods or class constructors',
@@ -208,9 +406,15 @@ class Ctor(CtorNames):
                     Ctor.ARGS: lambda x: [cls._serialize_ctor_to_jdict(x)],
                     Ctor.KWARGS: Literal(None),
                 },
-                'validate_conversion': lambda x, serialized_x: np.allclose(
-                    x, Ctor._construct_obj(serialized_x)
-                ),
+            },
+            {
+                'description': 'for functions',
+                'check_type': inspect.isfunction,
+                'spec': {
+                    Ctor.CONSTRUCTOR: Literal(cls._deserialize_ctor_from_jdict),
+                    Ctor.ARGS: lambda x: [cls._serialize_ctor_to_jdict(x)],
+                    Ctor.KWARGS: Literal(None),
+                },
             },
         ]
 
@@ -218,6 +422,15 @@ class Ctor(CtorNames):
     def construct(cls, obj):
         """Construct ctor objs. Return obj with no action if obj is str, int, float, bool, or None.
 
+        >>> from pprint import pprint
+        >>> from py2json.tests.test_ctor import add
+        >>> serialized = Ctor.deconstruct(add, validate_conversion=True, output_type=Ctor.CTOR_DICT)
+        >>> pprint(serialized, sort_dicts=False)
+        {'CONSTRUCTOR': <bound method Ctor._deserialize_ctor_from_jdict of <class 'ctor.Ctor'>>,
+         'ARGS': [{'module': 'py2json.tests.test_ctor', 'name': 'add', 'attr': None}],
+         'KWARGS': None}
+        >>> assert Ctor.construct(serialized) == add
+
         :param obj: ctor_dict or jdict
         :return: remap obj with constructed ctor_dicts
         """
@@ -256,6 +469,14 @@ class Ctor(CtorNames):
         """
         Remap a nested structure by deconstructing nodes of non-basic types into ctor_dicts using the deconstruction_specs
 
+        >>> from pprint import pprint
+        >>> from py2json.tests.test_ctor import add
+        >>> serialized = Ctor.deconstruct(add, validate_conversion=True, output_type=Ctor.CTOR_DICT)
+        >>> pprint(serialized, sort_dicts=False)
+        {'CONSTRUCTOR': <bound method Ctor._deserialize_ctor_from_jdict of <class 'ctor.Ctor'>>,
+         'ARGS': [{'module': 'py2json.tests.test_ctor', 'name': 'add', 'attr': None}],
+         'KWARGS': None}
+
         :param obj: any object containing only basic types and those described in the deconstruction_specs
         :param validate_conversion: [boolean] True to compare obj to reconstructed ctor_dict. False to skip validation
         :param output_type: Ctor.JSON_DICT or Ctor.CTOR_DICT
@@ -301,8 +522,8 @@ class Ctor(CtorNames):
 
     @classmethod
     def serializer(cls, ctor_pydict):
-        """
-        Search ctor_pydict and replace any Ctor.CONSTRUCTOR nodes with a JSON equivalent
+        """Search ctor_pydict and replace any Ctor.CONSTRUCTOR nodes with a JSON equivalent
+
         :param ctor_pydict: A ctor_dict or Dict containing nested ctor_dicts
         :return: jsonizeable ctor_jdict
         """
@@ -314,8 +535,7 @@ class Ctor(CtorNames):
 
     @classmethod
     def deserializer(cls, ctor_jdict):
-        """
-        Search ctor_jdict and replace any Ctor.CONSTRUCTOR nodes with a module object callable
+        """Search ctor_jdict and replace any Ctor.CONSTRUCTOR nodes with a module object callable
 
         :param ctor_jdict: A jsonized ctor_dict or Dict containing nested jsonized ctor_dicts
         :return:
@@ -327,7 +547,7 @@ class Ctor(CtorNames):
         )
 
     @classmethod
-    def to_jdict(cls, ctor_dict):
+    def to_jdict(cls, ctor_dict: CtorDict):
         return cls.serializer(ctor_dict)
 
     @classmethod
@@ -336,12 +556,45 @@ class Ctor(CtorNames):
         return cls._construct_obj(ctor_dict)
 
     @classmethod
-    def to_ctor_dict(cls, constructor: Callable, args: list, kwargs: dict):
-        return {
+    def to_ctor_dict(
+        cls,
+        constructor: Callable,
+        args: Union[list, tuple] = NOT_SET,
+        kwargs: dict = NOT_SET,
+        *,
+        output_type=CtorNames.JSON_DICT,
+    ) -> CtorDict:
+        """Function calls can also be serialized
+
+        >>> from pprint import pprint
+        >>> from collections import namedtuple
+        >>> ctor_jdict = Ctor.to_ctor_dict(namedtuple, ('A', 'x y z'), {'defaults': ('has', 'defaults')})
+        >>> pprint(ctor_jdict, sort_dicts=False)
+        {'CONSTRUCTOR': {'module': 'collections', 'name': 'namedtuple', 'attr': None},
+         'ARGS': ('A', 'x y z'),
+         'KWARGS': {'defaults': ('has', 'defaults')}}
+        >>> A = Ctor.construct(ctor_jdict)
+        >>> A('this one')
+        A(x='this one', y='has', z='defaults')
+
+        :param constructor: callable
+        :param args: args passed to callable
+        :param kwargs: kwargs passed to callable
+        :param output_type: JSON_DICT or CTOR_DICT
+        :return: CtorDict ready to be constructed with Ctor.construct()
+        """
+        if args is NOT_SET:
+            args = []
+        if kwargs is NOT_SET:
+            kwargs = {}
+        ctor_dict = {
             cls.CONSTRUCTOR: constructor,
-            cls.KWARGS: kwargs,
             cls.ARGS: args,
+            cls.KWARGS: kwargs,
         }
+        if output_type == CtorNames.JSON_DICT:
+            return cls.to_jdict(ctor_dict)
+        return ctor_dict
 
     # Boolean ..........................................................................................................
     @classmethod
@@ -377,7 +630,7 @@ class Ctor(CtorNames):
 
     # Private ..........................................................................................................
     @classmethod
-    def _construct_obj(cls, ctor_dict):
+    def _construct_obj(cls, ctor_dict: CtorDict):
         """
         Calls the Ctor.CONSTRUCTOR with given Ctor.ARGS and Ctor.KWARGS.
         Checks if ARGS or KWARGS are ctor_dicts and constructs those first
@@ -398,11 +651,13 @@ class Ctor(CtorNames):
             )
 
     @classmethod
-    def _deconstruct_obj(cls, obj, validate_conversion: bool = False):
-        """
-        Breakdown an obj into a ctor_dict as described by deconstruction_specs.
+    def _deconstruct_obj(cls, obj, validate_conversion: bool = False) -> CtorDict:
+        """Breakdown an obj into a ctor_dict as described by deconstruction_specs.
         Further breakdowns on deconstructed ARGS and KWARGS if necessary.
 
+        Will use fallback_deconstruction_spec to dill pickle object if no spec is found.
+        Validation is always performed in this fallback case.
+
         :param obj: any object satisfying one of the deconstruction_specs
         :param validate_conversion: [boolean] True to compare obj to reconstructed ctor_dict. False to skip validation
         :return: ctor_dict: {Ctor.CONSTRUCTOR: Callable, Ctor.ARGS: List[Any], Ctor.KWARGS: Dict[str, Any]}
@@ -414,6 +669,7 @@ class Ctor(CtorNames):
             except KeyError:
                 s['serializer'] = mk_serializer_and_deserializer(s['spec'])
                 serializer = s['serializer']
+
             ctor_dict = serializer(obj)
 
             if ctor_dict[Ctor.ARGS] is not None:
@@ -426,21 +682,33 @@ class Ctor(CtorNames):
                 )
 
             if validate_conversion is True:
-                try:
-                    validator = s['validate_conversion']
-                except KeyError:
-                    s['validate_conversion'] = cls._default_conversion_validation
-                    validator = s['validate_conversion']
-                if validator(obj, ctor_dict) is False:
-                    raise CtorException(
-                        f'deconstruction validation test failed for: {obj}'
-                    )
+                cls._validate_conversion(obj, ctor_dict, s)
 
             return ctor_dict
         except (StopIteration, KeyError):
-            raise CtorException(
-                f'deconstruction spec not found for non-basic type: {obj}'
-            )
+            try:
+                s = cls.fallback_deconstruction_spec
+                serializer = mk_serializer_and_deserializer(s['spec'])
+                ctor_dict = serializer(obj)
+                cls._validate_conversion(obj, ctor_dict, s)
+                return ctor_dict
+
+            except Exception as e:
+                raise CtorException(
+                    f'deconstruction spec not found for non-basic type: {obj}'
+                ) from e
+
+    @classmethod
+    def _validate_conversion(cls, obj, ctor_dict, deconstruction_spec):
+        try:
+            validator = deconstruction_spec['validate_conversion']
+        except KeyError:
+            deconstruction_spec[
+                'validate_conversion'
+            ] = cls._default_conversion_validation
+            validator = deconstruction_spec['validate_conversion']
+        if validator(obj, ctor_dict) is False:
+            raise CtorException(f'deconstruction validation test failed for: {obj}')
 
     @classmethod
     def _get_value(cls, dict_obj, key, default_value):
